#!/usr/bin/env python
import reflex 
import sys 
import os
import itertools
import json
import warnings
try: 
    from optparse import OptionParser
    from astropy.io import fits
    import numpy as np
    from numpy import array
    from photutils import datasets
    from photutils import daofind
    from astropy.io import fits as pyfits
    from astropy import wcs as pywcs
    import astropy
    import scipy
    import copy
    import matplotlib.pyplot as plt
    import_ok = 1
except ImportError:
    import_ok = 0
def offsets(catalogs,sources):

#********************************************************************************************
# parameters
#********************************************************************************************

# search radius for each iteration in arcsec.
# The first radius has to be bigger than the maximum offset. 
# the last radius should be approximately the size of the 
# random erros in coordinates:
 search_radii=[30,4,2] 


# number of bins for 2-dimensional mode:
 nbin=30

# switch weighting on or off:
 do_weight= True

#**********************************************************************************************
# read in catalogs
#**********************************************************************************************

# this is a list of tupels [ra,dec,field_id], where ra and dec are
# vectors, and field_id is a running number give to the field.
 fieldlist= []  
 for icat,catalog in enumerate(catalogs):
        ncat= icat+1  # number of catalogs
        ra = sources[icat][1]
        dec = sources[icat][2]
        field_id= icat
        fieldlist.append([ra,dec,field_id])





#*********************************************************************************************
# iterate over different serach radii:
#*********************************************************************************************

 for iiter,search_radius in enumerate(search_radii):
        #*************************************************************************************
        # The goal is to measure the offset in measured_offsets_ra and
        # measured_offsets_dec between any two fields, and then write these measured
        # offsets into a set of equations that can be solved for the offset
        # offset_ra and offset_dec  to be applied to each field.
        #
        # These equations are of the form 
        #
        #       equations*offset_ra=measured_offsets_ra
        #       equations*offset_dec=measured_offsets_dec
        #
        # where "equations" is a matrix, offset_ra and offset_dec are the
        # offset vector (one number for each field), and measured_offsets_ra and
        # measured_offsets_dec are the measured offsets between any two images. 
        # If all fields overlap, the length of measured_offsets_* is ncat*(ncat-1)/2, 
        # and the dimension of the matrix equations is ncat x  ncat*(ncat-1)/2.:w
        #
        # Each equation (i.e. any row in the matrix and the corresponding element in 
        # measured offset) can be multiplied by a weight.
        # 
        # The task here is to construct the matrix "equations" and the measured_offsets_*:
        #*************************************************************************************
        
        # The first equations is to set the first  offset to zero:
        equations= np.zeros(ncat) 
        equations[0]= 1 
        measured_offsets_ra=  np.array([0])
        measured_offsets_dec= np.array([0]) 

        # now loop for any combination of fields:
        for field1,field2 in  itertools.combinations(fieldlist,2): 

                # recover the star coordinates of the two fields: 
                ra1=       field1[0] ;  ra2=       field2[0] 
                dec1=      field1[1] ;  dec2=      field2[1] 
                field_id1= field1[2] ;  field_id2= field2[2] 

        
                # compute relative distance between any star in field 1 to any star in field 2,
                # using small angle approximation. Use matrix operation to avoid loops:
                delta_ra = (np.array([ra1,]*len(ra2)).transpose() - np.array([ra2,]*len(ra1))) *3600
                delta_dec= (np.array([dec1,]*len(dec2)).transpose() - np.array([dec2,]*len(dec1)))*3600

                # compute mean declination (in radians) for any pair of stars:
                dec_rad= (np.array([dec1,]*len(dec2)).transpose() + np.array([dec2,]*len(dec1)))/2. * np.pi/180.
        
                if iiter == 0:
                        #*********************************************************************
                        # In the first iteration, there are no estimates for offsets yet.
                        # Therefore we use the 2-dimensional mode to estimates offsets:
                        #*********************************************************************

                        # compute distance between any pair of stars using small angle approximation:
                        dist= np.sqrt((delta_ra*np.cos(dec_rad))**2 + delta_dec**2)

                        # check whether there is any overlap between the two list of stars: 
                        if (np.min(dist) > search_radius):
                                # there is no overlap, lets try the next two fields:
                                continue

                        # throw out stars further apart than the search radius. This is 
                        # needed to avoid passing numbers outside of the bins to
                        # the histogram2d routine:
                        candidates= (dist<search_radius)
                        delta_ra= delta_ra[candidates]
                        delta_dec= delta_dec[candidates]

                        # compute mode of 2d distribution (delta_ra, delta_dec) using the
                        # 2d histogram function:
                        h, ra_bin, dec_bin=np.histogram2d(delta_ra, delta_dec, bins=nbin, range=[[-search_radius,search_radius],[-search_radius,search_radius]])
                        ra_max_ind= np.unravel_index(np.ndarray.argmax(h), [nbin,nbin])
                        ra_offset12= ( ra_bin[ra_max_ind[0]] +ra_bin[ra_max_ind[0]+1])/2.
                        dec_offset12= ( dec_bin[ra_max_ind[1]] +dec_bin[ra_max_ind[1]+1])/2.

                        # weight for this equation is the height of the peak in the histogram:
                        if do_weight:
                                weight= np.max(h)
                        else:  
                                weight= 1.

                else :
                        #*********************************************************************
                        # In subsequent iterations, there are already estimates for offsets.
                        # Therefore we use these estimates to select stars that are in more
                        # than one field, and simply take the median of the measured 
                        # delta_ra and delta_dec
                        #*********************************************************************

                        # compute distance from any star on field 1 to any star in field 2,
                        # taking previously determined offsets of the two fields into account. 
                        # This is only used to select stars around expected positions, so small 
                        #angle approximation will do:
                        dist= np.sqrt( ((delta_ra-offset_ra[field_id1]+offset_ra[field_id2])*np.cos(dec_rad))**2+ \
                                        (delta_dec-offset_dec[field_id1]+offset_dec[field_id2])**2 )

                        # check whether there is any overlap between the two list of stars: 
                        if (np.min(dist) > search_radius):
                                # there is no overlap, lets try the next two fields:
                                continue

                        # throw out stars further apart than the search radius. This is 
                        # essential to compute a meaningful median:
                        candidates= (dist<search_radius)
                        delta_ra= delta_ra[candidates]
                        delta_dec= delta_dec[candidates]

                        # estimate offsets as the median of all delta_* within search radius:
                        ra_offset12 = np.median(delta_ra)
                        dec_offset12= np.median(delta_dec)

                        
                        # the weights in this case are the inverse of the square of the uncertainy 
                        # of the mean measured offsets. 
                        # This ingores contamination by interloopers:
                        if do_weight:
                                if len(delta_ra) == 1:
                                        # if there is only one offset, we assume the uncertainty 
                                        # to be the search radius:
                                        var= 2*(search_radius)**2
                                else:
                                        var= np.var(delta_ra)+np.var(delta_dec)
                                weight= len(delta_ra)/var
                        else:  
                                weight= 1.
                        
                # now let's write the equation for this pair of fields. 
                # Note the use of the weight:
                equation= np.zeros(ncat)
                equation[field_id1]= weight
                equation[field_id2]= -weight
                measured_offset_ra= [weight*ra_offset12]
                measured_offset_dec= [weight*dec_offset12]

                # add this equation to the matrix and vectors:
                equations= np.vstack((equations, equation))
                measured_offsets_ra= np.concatenate((measured_offsets_ra, measured_offset_ra))
                measured_offsets_dec= np.concatenate((measured_offsets_dec, measured_offset_dec))
        
        #*************************************************************************************
        # finally, we have all the equations written so it can be solved with a least
        # square equation solver.
        #*************************************************************************************
        offset_ra= np.linalg.lstsq(equations, measured_offsets_ra)[0]
        offset_dec= np.linalg.lstsq(equations, measured_offsets_dec)[0]




        
 # conversion to degrees.
 offset_ra = offset_ra/3600.
 offset_dec = offset_dec/3600.
        
 return(offset_ra,offset_dec)






if __name__ == '__main__':

  
#   MAIN MODULE ---
  warnings.filterwarnings("ignore")
 #getting the inputs and defyning the outputs
  parser = reflex.ReflexIOParser()
  parser.add_option("-i", "--in_sof", dest="in_sof")
  parser.add_output("-o", "--out_sof", dest="out_sof")
  parser.add_output("-p", "--messages", dest="messages")
  inputs  = parser.get_inputs()
  outputs = parser.get_outputs()
  in_sof = inputs.in_sof
  files=in_sof.files
  images=[]
  tables=[]
  if import_ok == 0: # in the case import fails
      outputs.out_sof = inputs.in_sof
      outputs.messages = ('Warning: Failing in importing some required Python modules.\n Check software requirements at:\n '
            ' https://www.eso.org/sci/software/pipelines/reflex_workflows/#software_prerequisites\n'
            ' or in the MUSE Reflex tutorial.\n\n '
            'Coordinates in the header will be used for alignment.\n')
      parser.write_outputs()
      sys.exit()
 
  for file in files:
 
      if file.category == 'IMAGE_FOV':
          images.append(str(file.name))
      if file.category == 'PIXTABLE_REDUCED':    
          tables.append(str(file.name))

# SORTING THE IMAGES IN TIME
  sources = []
  RA=np.zeros(len(images),float)
  DEC=np.zeros(len(images),float)
  MJD=np.zeros(len(images),float)
  MJD_tbl=np.zeros(len(images),float)
  FLAGS=np.zeros(len(images),float)
  FLAGS[0] = 1.
  OFFSETS_RA=np.zeros(len(images),float)
  OFFSETS_DEC=np.zeros(len(images),float)
  distances=np.zeros([len(images),len(images)],float)

  for i in range(len(images)):
      hdu  = pyfits.open(images[i])
      image = hdu[1].data 
  
      # REPLACE NaN values with 0
      where_are_NaNs = np.isnan(image)
      image[where_are_NaNs] = 0.
  
      header = hdu[0] 
      RA[i] = header.header['RA']
      DEC[i] = header.header['DEC']
      MJD[i] = header.header['MJD-OBS']
      wcs = pywcs.WCS(hdu[0].header)

      hdu_tbl  = pyfits.open(tables[i])
      header_tbl = hdu_tbl[0] 
      MJD_tbl[i] = header_tbl.header['MJD-OBS']


  sorted=np.argsort(MJD)
  sorted_tbl=np.argsort(MJD_tbl)
  RA_sorted=RA[sorted]
  DEC_sorted=DEC[sorted]
  MJD_sorted=MJD[sorted]
  MJD_tbl_sorted=MJD_tbl[sorted_tbl]
  images_array = np.array(images)
  tables_array=np.array(tables)
  images_array_sorted=images_array[sorted]
  tables_array_sorted=tables_array[sorted_tbl]
  images_sorted = np.ndarray.tolist(images_array_sorted)
  tables_sorted = np.ndarray.tolist(tables_array_sorted)
 

# FINDING THE SOURCES
  for i in range(len(images_sorted)):
      hdu  = pyfits.open(images_sorted[i])
      image = hdu[1].data 
  
      # REPLACE NaN values with 0
      where_are_NaNs = np.isnan(image)
      image[where_are_NaNs] = 0.
      wcs = pywcs.WCS(hdu[0].header)

      thr = 15.
      try:
          junk =  daofind(image, fwhm=5.0, threshold=thr,exclude_border=True) 
          len_junk = len(junk)
          while len_junk > 80:    # I do not want more than 80 sources to avouid crowdking problems
             thr=thr+5.
             junk =  daofind(image, fwhm=5.0, threshold=thr,exclude_border=True) 
             len_junk = len(junk)
          # decreasing the threshold if less than 5 sources
          if len(junk) <= 5: 
             junk =  daofind(image, fwhm=5.0, threshold=10.,exclude_border=True) 
          if len(junk) <= 5: 
             junk =  daofind(image, fwhm=5.0, threshold=5.,exclude_border=True) 
          #if it still fails and an error is raised, the except KeyError case will be executed.
         
          x=np.array(junk['xcentroid'])
          y=np.array(junk['ycentroid'])
          id_=np.array(junk['id'])
          sky = wcs.wcs_pix2world(x,y,1)
          sources.append([id_,sky[0],sky[1]])
      except KeyError:
          sources.append([np.array([-1]),np.array([-999]),np.array([-999])])


  # FINDING THE OFFSETS
  offsets_to_the_reference_RA,offsets_to_the_reference_DEC = offsets(images_sorted,sources)

  # ADJUSTING THE HEADERS OF IMAGES AND PIXTABLES TO ACCOUNT FOR THE OFFSETS:
  for i in range(len(images_sorted)):
          
      hdu = pyfits.open(tables_sorted[i],mode='update')
      hdu[0].header['RA'] = hdu[0].header['RA'] - offsets_to_the_reference_RA[i]
      hdu[0].header['DEC'] = hdu[0].header['DEC'] - offsets_to_the_reference_DEC[i]
      hdu[0].header['HIERARCH REFLEX ALIGNED'] = 1
      hdu[0].header['HIERARCH REFLEX APPOFFRA'] = (offsets_to_the_reference_RA[i],'Offset in RA (deg)')
      hdu[0].header['HIERARCH REFLEX APPOFFDEC'] = (offsets_to_the_reference_DEC[i],'Offset in DEC (deg)')
      hdu.flush() 

      hdu = pyfits.open(images_sorted[i],mode='update')
      hdu[0].header['CRVAL1'] = hdu[0].header['CRVAL1'] - offsets_to_the_reference_RA[i]
      hdu[0].header['CRVAL2'] = hdu[0].header['CRVAL2'] - offsets_to_the_reference_DEC[i]
      hdu[0].header['RA'] = hdu[0].header['RA'] - offsets_to_the_reference_RA[i]
      hdu[0].header['DEC'] = hdu[0].header['DEC'] - offsets_to_the_reference_DEC[i]
      hdu[0].header['HIERARCH REFLEX ALIGNED'] = 1
      hdu[0].header['HIERARCH REFLEX APPOFFRA'] = (offsets_to_the_reference_RA[i],'Offset in RA (deg)')
      hdu[0].header['HIERARCH REFLEX APPOFFDEC'] = (offsets_to_the_reference_DEC[i],'Offset in DEC (deg)')
      hdu.flush()
# 
  outputs.out_sof = inputs.in_sof
  parser.write_outputs()
  sys.exit()
 


